#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.channels_name as channels_name
import octobot_commons.enums as commons_enums
import octobot_commons.authentication as authentication
import octobot_commons.databases as commons_databases

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.storage.abstract_storage as abstract_storage
import octobot_trading.storage.util as storage_util


class TradesStorage(abstract_storage.AbstractStorage):
    LIVE_CHANNEL = channels_name.OctoBotTradingChannelsName.TRADES_CHANNEL.value
    HISTORY_TABLE = commons_enums.DBTables.TRADES.value

    async def _live_callback(
        self,
        exchange: str,
        exchange_id: str,
        cryptocurrency: str,
        symbol: str,
        trade: dict,
        old_trade: bool
    ):
        if trade[enums.ExchangeConstantsOrderColumns.STATUS.value] != enums.OrderStatus.CANCELED.value:
            await self._get_db().log(
                self.HISTORY_TABLE,
                _format_trade(
                    trade,
                    self.exchange_manager,
                    self.plot_settings.chart,
                    self.plot_settings.x_multiplier,
                    self.plot_settings.kind,
                    self.plot_settings.mode
                )
            )
            await self.trigger_debounced_flush()
            self._to_update_auth_data_ids_buffer.add(trade[enums.ExchangeConstantsOrderColumns.ID.value])
            await self.trigger_debounced_update_auth_data(False)

    async def _update_auth_data(self, reset):
        authenticator = authentication.Authenticator.instance()
        history = [
            trade.to_dict()
            for trade in self.exchange_manager.exchange_personal_data.trades_manager.trades.values()
            if trade.status is not enums.OrderStatus.CANCELED and trade.trade_id in self._to_update_auth_data_ids_buffer
        ]
        if (history or reset) and authenticator.is_initialized():
            # also update when history is empty to reset trade history
            await authenticator.update_trades(history, self.exchange_manager.exchange_name, reset)
            self._to_update_auth_data_ids_buffer.clear()

    async def _store_history(self):
        database = self._get_db()
        await database.replace_all(
            self.HISTORY_TABLE,
            [
                _format_trade(
                    trade.to_dict(),
                    self.exchange_manager,
                    self.plot_settings.chart,
                    self.plot_settings.x_multiplier,
                    self.plot_settings.kind,
                    self.plot_settings.mode
                )
                for trade in self.exchange_manager.exchange_personal_data.trades_manager.trades.values()
                if trade.status is not enums.OrderStatus.CANCELED
            ],
            cache=False,
        )
        await database.flush()

    def _get_db(self):
        return commons_databases.RunDatabasesProvider.instance().get_trades_db(
            self.exchange_manager.bot_id,
            storage_util.get_account_type_suffix_from_exchange_manager(self.exchange_manager),
            self.exchange_manager.exchange_name,
        )


def _format_trade(trade_dict, exchange_manager, chart, x_multiplier, kind, mode):
    tag = f"{trade_dict[enums.ExchangeConstantsOrderColumns.TAG.value]} " \
        if trade_dict[enums.ExchangeConstantsOrderColumns.TAG.value] else ""
    symbol = trade_dict[enums.ExchangeConstantsOrderColumns.SYMBOL.value]
    trade_side = trade_dict[enums.ExchangeConstantsOrderColumns.SIDE.value]
    is_using_positions = False
    color = shape = None
    if exchange_manager.is_future:
        positions = exchange_manager.exchange_personal_data.positions_manager.get_symbol_positions(symbol=symbol)
        if positions:
            is_using_positions = True
            # trading_side = next(iter(positions)).side
            # if trading_side is enums.PositionSide.LONG:
            if "stop_loss" in trade_dict[enums.ExchangeConstantsOrderColumns.TYPE.value]:
                shape = "x"
                color = "orange"
            elif trade_dict[enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value] is True:
                if trade_side == enums.TradeOrderSide.SELL.value:
                    # long tp
                    color = "magenta"
                    shape = "arrow-bar-left"
                else:
                    # short tp
                    color = "blue"
                    shape = "arrow-bar-left"
            else:
                if trade_side == enums.TradeOrderSide.BUY.value:
                    # long entry
                    color = "green"
                    shape = "arrow-bar-right"
                else:
                    # short entry
                    color = "red"
                    shape = "arrow-bar-right"

    if not is_using_positions:
        if trade_side == enums.TradeOrderSide.BUY.value:
            color = "blue"
            shape = "arrow-bar-right"
        elif "stop_loss" in trade_dict[enums.ExchangeConstantsOrderColumns.TYPE.value]:
            color = "orange"
            shape = "x"
        else:
            color = "magenta"
            shape = "arrow-bar-left"
    fee = trade_dict[enums.ExchangeConstantsOrderColumns.FEE.value]
    fee_cost = float(fee[enums.FeePropertyColumns.COST.value] if
                     fee and fee[enums.FeePropertyColumns.COST.value] else 0)
    return {
        constants.STORAGE_ORIGIN_VALUE: TradesStorage.sanitize_for_storage(trade_dict),
        commons_enums.DisplayedElementTypes.CHART.value: chart,
        commons_enums.DBRows.SYMBOL.value: trade_dict[enums.ExchangeConstantsOrderColumns.SYMBOL.value],
        commons_enums.DBRows.FEES_AMOUNT.value: fee_cost,
        commons_enums.DBRows.FEES_CURRENCY.value: fee[enums.FeePropertyColumns.CURRENCY.value]
        if trade_dict[enums.ExchangeConstantsOrderColumns.FEE.value] else "",
        commons_enums.DBRows.ID.value: trade_dict[enums.ExchangeConstantsOrderColumns.ID.value],
        commons_enums.DBRows.TRADING_MODE.value: exchange_manager.trading_modes[0].get_name(),
        commons_enums.PlotAttributes.X.value: trade_dict[enums.ExchangeConstantsOrderColumns.TIMESTAMP.value] * x_multiplier,
        commons_enums.PlotAttributes.TEXT.value: f"{tag}{trade_dict[enums.ExchangeConstantsOrderColumns.TYPE.value]} "
                f"{trade_dict[enums.ExchangeConstantsOrderColumns.SIDE.value]} "
                f"{trade_dict[enums.ExchangeConstantsOrderColumns.AMOUNT.value]} "
                f"{trade_dict[enums.ExchangeConstantsOrderColumns.QUANTITY_CURRENCY.value]} "
                f"at {trade_dict[enums.ExchangeConstantsOrderColumns.PRICE.value]}",
        commons_enums.PlotAttributes.TYPE.value: trade_dict[enums.ExchangeConstantsOrderColumns.TYPE.value],
        commons_enums.PlotAttributes.VOLUME.value: float(trade_dict[enums.ExchangeConstantsOrderColumns.AMOUNT.value]),
        commons_enums.PlotAttributes.Y.value: float(trade_dict[enums.ExchangeConstantsOrderColumns.PRICE.value]),
        commons_enums.PlotAttributes.KIND.value: kind,
        commons_enums.PlotAttributes.SIDE.value: trade_dict[enums.ExchangeConstantsOrderColumns.SIDE.value],
        commons_enums.PlotAttributes.MODE.value: mode,
        commons_enums.PlotAttributes.SHAPE.value: shape,
        commons_enums.PlotAttributes.COLOR.value: color,
        commons_enums.PlotAttributes.SIZE.value: "10",
        "cost": float(trade_dict[enums.ExchangeConstantsOrderColumns.COST.value]),
        "state": trade_dict[enums.ExchangeConstantsOrderColumns.STATUS.value],
    }



# Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.constants as commons_constants
import octobot_tentacles_manager.api as api
import octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket


def force_disable_web_socket(config, exchange_name) -> bool:
    return commons_constants.CONFIG_EXCHANGE_WEB_SOCKET in config[commons_constants.CONFIG_EXCHANGES][exchange_name] \
           and not config[commons_constants.CONFIG_EXCHANGES][exchange_name][commons_constants.CONFIG_EXCHANGE_WEB_SOCKET]


def check_web_socket_config(config, exchange_name) -> bool:
    return not force_disable_web_socket(config, exchange_name)


def search_websocket_class(websocket_class, exchange_manager):
    for socket_manager in tentacles_management.get_all_classes_from_parent(websocket_class):
        # return websocket exchange if available
        if socket_manager.has_name(exchange_manager):
            return socket_manager
    return None


def supports_websocket(exchange_name, tentacles_setup_config) -> bool:
    for connector in abstract_websocket.AbstractWebsocketExchange.__subclasses__():
        try:
            if api.get_class_from_name_with_activated_required_tentacles(
                name=exchange_name,
                tentacles_setup_config=tentacles_setup_config,
                parent_class=connector
            ) is not None:
                return True
        except NotImplementedError:
            pass
    return False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.types.trailing.trailing_stop_order as trailing_stop_order


class TrailingStopLimitOrder(trailing_stop_order.TrailingStopOrder):
    UNINITIALIZED_LIMIT_PRICE = -1

    def __init__(self, trader, side=enums.TradeOrderSide.SELL, limit_price=UNINITIALIZED_LIMIT_PRICE):
        super().__init__(trader, side)
        self.order_type = enums.TraderOrderType.TRAILING_STOP_LIMIT
        self.limit_price = limit_price

    async def on_filled(self):
        await trailing_stop_order.TrailingStopOrder.on_filled(self)
        # TODO replace with chained order ?
        await self.trader.create_artificial_order(enums.TraderOrderType.SELL_LIMIT
                                                  if self.side is enums.TradeOrderSide.SELL else enums.TraderOrderType.BUY_LIMIT,
                                                  self.symbol, self.origin_stop_price,
                                                  self.origin_quantity,
                                                  self.limit_price
                                                  if self.limit_price != self.UNINITIALIZED_LIMIT_PRICE else
                                                  self.origin_stop_price)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import decimal

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data as trading_personal_data
import octobot_trading.modes.script_keywords as script_keywords
import octobot_trading.modes.script_keywords.dsl as dsl
import octobot_trading.modes.script_keywords.basic_keywords.account_balance as account_balance

from tests import event_loop
from tests.modes.script_keywords import null_context

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_get_amount_from_input_amount(null_context):
    with pytest.raises(errors.InvalidArgumentError):
        await script_keywords.get_amount_from_input_amount(null_context, "-1")

    with pytest.raises(errors.InvalidArgumentError):
        await script_keywords.get_amount_from_input_amount(null_context, "1sdsqdq")

    with pytest.raises(NotImplementedError):
        await script_keywords.get_amount_from_input_amount(null_context,
                                                           f"1{script_keywords.QuantityType.POSITION_PERCENT.value}")

    with mock.patch.object(account_balance, "adapt_amount_to_holdings",
                           mock.AsyncMock(return_value=decimal.Decimal(1))) as adapt_amount_to_holdings_mock:
        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.DELTA, decimal.Decimal(2)))) \
                as parse_quantity_mock:
            assert await script_keywords.get_amount_from_input_amount(null_context, "1", "buy") == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal(2), "buy",
                                                                  False, True, False, target_price=None)
            parse_quantity_mock.assert_called_once_with("1")
            adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.DELTA_QUOTE, decimal.Decimal(2)))) \
                as parse_quantity_mock:
            with mock.patch.object(trading_personal_data, "get_up_to_date_price",
                                   mock.AsyncMock(return_value=decimal.Decimal(100))) \
                    as get_up_to_date_price_mock:
                assert await script_keywords.get_amount_from_input_amount(null_context, "1", "buy") == decimal.Decimal(1)
                get_up_to_date_price_mock.assert_awaited_with(
                    null_context.exchange_manager,
                    symbol=null_context.symbol,
                    timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
                )
                # without target_price
                adapt_amount_to_holdings_mock.assert_called_once_with(
                    null_context, decimal.Decimal(2) / decimal.Decimal(100), "buy",
                    False, True, False, target_price=None
                )
                parse_quantity_mock.assert_called_once_with("1")
                adapt_amount_to_holdings_mock.reset_mock()
                parse_quantity_mock.reset_mock()
                get_up_to_date_price_mock.reset_mock()
                # with target_price
                assert await script_keywords.get_amount_from_input_amount(
                    null_context, "1", "buy", target_price=decimal.Decimal(23)
                ) == decimal.Decimal(1)
                get_up_to_date_price_mock.assert_not_called()
                adapt_amount_to_holdings_mock.assert_called_once_with(
                    null_context, decimal.Decimal(2) / decimal.Decimal(23), "buy",
                    False, True, False, target_price=decimal.Decimal(23)
                )
                parse_quantity_mock.assert_called_once_with("1")
                adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(script_keywords.QuantityType.PERCENT, decimal.Decimal(75)))) \
                as parse_quantity_mock, \
                mock.patch.object(account_balance, "available_account_balance",
                                  mock.AsyncMock(return_value=decimal.Decimal(2))) \
                as available_account_balance_mock:
            assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy", use_total_holding=True,
                                                                      reduce_only=False,
                                                                      is_stop_order=True,
                                                                      target_price=constants.ZERO) == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("1.5"), "buy",
                                                                  True, False, True, target_price=constants.ZERO)
            parse_quantity_mock.assert_called_once_with("50")
            available_account_balance_mock.assert_called_once_with(null_context, "buy",
                                                                   use_total_holding=True, reduce_only=False)
            adapt_amount_to_holdings_mock.reset_mock()

        with mock.patch.object(dsl, "parse_quantity",
                               mock.Mock(return_value=(
                                       script_keywords.QuantityType.AVAILABLE_PERCENT, decimal.Decimal(75)))) \
                as parse_quantity_mock, \
                mock.patch.object(account_balance, "available_account_balance",
                                  mock.AsyncMock(return_value=decimal.Decimal(2))) \
                as available_account_balance_mock:
            assert await script_keywords.get_amount_from_input_amount(null_context, "50", "buy") == decimal.Decimal(1)
            adapt_amount_to_holdings_mock.assert_called_once_with(null_context, decimal.Decimal("1.5"), "buy",
                                                                  False, True, False, target_price=None)
            parse_quantity_mock.assert_called_once_with("50")
            available_account_balance_mock.assert_called_once_with(null_context, "buy",
                                                                   use_total_holding=False, reduce_only=True)
            adapt_amount_to_holdings_mock.reset_mock()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import enum


class ExchangeConstantsCCXTColumns(enum.Enum):
    TIMESTAMP = "timestamp"
    DATETIME = "datetime"


class ExchangeConstantsMarketStatusCCXTColumns(enum.Enum):
    CONTRACT_SIZE = "contractSize"


class ExchangePositionCCXTColumns(enum.Enum):
    CONTRACTS = "contracts"
    CONTRACT_SIZE = "contractSize"
    MARGIN_TYPE = "marginType"
    MARGIN_MODE = "marginMode"
    LEVERAGE = "leverage"
    SYMBOL = "symbol"
    COLLATERAL = "collateral"
    INITIAL_MARGIN = "initialMargin"
    INITIAL_MARGIN_PERCENTAGE = "initialMarginPercentage"
    MAINTENANCE_MARGIN = "maintenanceMargin"
    MAINTENANCE_MARGIN_PERCENTAGE = "maintenanceMarginPercentage"
    NOTIONAL = "notional"
    MARGIN_RATIO = "marginRatio"
    UNREALISED_PNL = "unrealizedPnl"
    REALISED_PNL = "realizedPnl"
    LIQUIDATION_PRICE = "liquidationPrice"
    MARK_PRICE = "markPrice"
    ENTRY_PRICE = "entryPrice"
    TIMESTAMP = "timestamp"
    DATETIME = "datetime"
    PERCENTAGE = "percentage"
    SIDE = "side"
    HEDGED = "hedged"
    INFO = "info"


class ExchangeFundingCCXTColumns(enum.Enum):
    SYMBOL = "symbol"
    LAST_FUNDING_TIME = "lastFundingTime"
    FUNDING_RATE = "fundingRate"
    FUNDING_TIMESTAMP = "fundingTimestamp"
    NEXT_FUNDING_TIME = "nextFundingTime"
    NEXT_FUNDING_TIMESTAMP = "nextFundingTimestamp"
    PREDICTED_FUNDING_RATE = "predictedFundingRate"
    PREVIOUS_FUNDING_TIMESTAMP = "previousFundingTimestamp"
    PREVIOUS_FUNDING_RATE = "previousFundingRate"


class ExchangeLeverageTiersCCXTColumns(enum.Enum):
    TIER = "tier"
    CURRENCY = "currency"
    MIN_NOTIONAL = "minNotional"
    MAX_NOTIONAL = "maxNotional"
    MAINTENANCE_MARGIN_RATE = "maintenanceMarginRate"
    MAX_LEVERAGE = "maxLeverage"
    INFO = "info"


class ExchangeOrderCCXTColumns(enum.Enum):
    INFO = "info"
    ID = "id"
    TIMESTAMP = "timestamp"
    DATETIME = 'datetime'
    LAST_TRADE_TIMESTAMP = "lastTradeTimestamp"
    SYMBOL = "symbol"
    QUANTITY_CURRENCY = "quantityCurrency"
    TYPE = "type"
    SIDE = "side"
    PRICE = "price"
    AMOUNT = "amount"
    COST = "cost"
    AVERAGE = "average"
    FILLED = "filled"
    REMAINING = "remaining"
    STATUS = "status"
    FEE = "fee"
    TRADES = "trades"
    MAKER = "maker"
    TAKER = "taker"
    ORDER = "order"
    TAKER_OR_MAKER = "takerOrMaker"
    REDUCE_ONLY = "reduceOnly"
    STOP_PRICE = "stopPrice"
    TRIGGER_PRICE = "triggerPrice"
    TRIGGER_ABOVE = "triggerAbove"
    TAG = "tag"
    MARGIN_MODE = "marginMode"


class ExchangeWrapperLibs(enum.Enum):
    ASYNC_CCXT = "async_ccxt"
    CCXT = "ccxt"


class ExchangeColumns(enum.Enum):
    WEBSITE = "www"
    URLS = "urls"
    API = "api"
    REST = "rest"
    WEBSOCKET = "ws"
    LOGO_URL = "logo"


class ExchangeMarginTypes(enum.Enum):
    ISOLATED = "isolated"
    CROSS = "cross"



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest

from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_trading.enums import TraderOrderType
import octobot_trading.constants as trading_constants

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data import DEFAULT_ORDER_SYMBOL, DEFAULT_SYMBOL_QUANTITY
from tests.personal_data.orders import sell_limit_order
from tests.test_utils.random_numbers import decimal_random_price, decimal_random_quantity, \
    decimal_random_recent_trade

pytestmark = pytest.mark.asyncio


async def test_sell_limit_order_trigger(sell_limit_order):
    order_price = decimal_random_price(min_value=2)
    sell_limit_order.update(
        price=order_price,
        quantity=decimal_random_quantity(max_value=DEFAULT_SYMBOL_QUANTITY),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.SELL_LIMIT,
    )
    sell_limit_order.exchange_manager.is_backtesting = True  # force update_order_status
    await sell_limit_order.initialize()
    await sell_limit_order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(
        sell_limit_order
    )
    price_events_manager = sell_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=decimal_random_price(max_value=order_price - trading_constants.ONE),
                                     timestamp=sell_limit_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not sell_limit_order.is_filled()
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=order_price,
                                     timestamp=sell_limit_order.timestamp - 1)])
    await wait_asyncio_next_cycle()
    assert not sell_limit_order.is_filled()
    price_events_manager.handle_recent_trades([decimal_random_recent_trade(price=order_price,
                                                                           timestamp=sell_limit_order.timestamp)])

    await wait_asyncio_next_cycle()
    assert sell_limit_order.is_filled()


async def test_sell_limit_order_on_origin_price_change(sell_limit_order):
    order_price = decimal.Decimal(100)
    sell_limit_order.update(
        price=order_price,
        quantity=decimal_random_quantity(max_value=DEFAULT_SYMBOL_QUANTITY),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.SELL_LIMIT,
    )
    sell_limit_order.exchange_manager.is_backtesting = True  # force update_order_status
    await sell_limit_order.initialize()
    await sell_limit_order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(
        sell_limit_order
    )
    price_events_manager = sell_limit_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=decimal_random_price(max_value=order_price - trading_constants.ONE),
                                     timestamp=sell_limit_order.timestamp)])
    await wait_asyncio_next_cycle()
    assert not sell_limit_order.is_filled()

    # do not update order price
    order_price = decimal.Decimal(10)
    sell_limit_order.update(
        symbol=sell_limit_order.symbol,
        quantity=decimal_random_quantity(max_value=DEFAULT_SYMBOL_QUANTITY)
    )
    price_events_manager.handle_recent_trades(
        [decimal_random_recent_trade(price=decimal_random_price(max_value=order_price - trading_constants.ONE),
                                     timestamp=sell_limit_order.timestamp)])
    # order is still not triggered by this price
    await wait_asyncio_next_cycle()
    assert not sell_limit_order.is_filled()

    # update order price
    order_price = decimal.Decimal(10)
    sell_limit_order.update(
        symbol=sell_limit_order.symbol,
        price=order_price
    )
    price_events_manager.handle_recent_trades([
        decimal_random_recent_trade(
            price=decimal.Decimal(20),
            timestamp=sell_limit_order.exchange_manager.exchange.get_exchange_current_time()
        )
    ])
    # order is now triggered by the new price
    await wait_asyncio_next_cycle()
    assert sell_limit_order.is_filled()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.enums as enums
import octobot_trading.personal_data.positions.position_state as position_state
import octobot_trading.personal_data.positions.states.position_state_factory as position_state_factory


class IdlePositionState(position_state.PositionState):
    """
    ActivePositionState is the state of a position that has size of zero
    """
    def __init__(self, position, is_from_exchange_data):
        super().__init__(position, is_from_exchange_data)
        self.state = enums.States.OPEN if is_from_exchange_data \
                                          or self.position.simulated \
                                          or self.position.status is enums.PositionStatus.OPEN \
            else enums.States.OPENING

        self.has_terminated = False

    def is_open(self) -> bool:
        """
        :return: True if the Position is considered as open
        """
        return not (self.is_pending() or self.is_refreshing())

    async def initialize_impl(self, forced=False) -> None:
        if forced:
            self.state = enums.States.OPEN
        return await super().initialize_impl()

    def _is_compatible_size(self):
        return self.position.is_idle()

    async def on_refresh_successful(self):
        """
        Verify the position is properly created and still PositionStatus.OPEN
        """
        # skip refresh process if the current position state is not the same as the one triggering this
        # on_refresh_successful to avoid synchronization issues (state already got refreshed by another mean)
        if self.position is None:
            self.get_logger().warning(f"on_refresh_successful triggered on cleared position: ignoring update.")
        elif self.state is self.position.state.state:
            if self.position.status is enums.PositionStatus.OPEN and self._is_compatible_size():
                self.state = enums.States.OPEN
                await self.update()
            else:
                position_state_factory.create_position_state(self.position, is_from_exchange_data=True)
                if self.position.state.has_to_be_async_synchronized():
                    await self.position.state.initialize()
        else:
            self.get_logger().debug(f"on_refresh_successful triggered from previous state "
                                    f"after state change on {self.position}")

    async def _synchronize_with_exchange(self, force_synchronization: bool = False) -> None:
        # Disable open position synchronization for now, let position updater refresh positions
        pass

    def sync_terminate(self):
        """
        Should wait for being replaced by an ActivePositionState or LiquidatePositionState
        """
        if not self.has_terminated:
            self.log_event_message(enums.StatesMessages.OPEN)

            if not self.position.mark_price:
                # set position mark price
                self.position.mark_price = decimal.Decimal(
                    self.position.exchange_manager.exchange_symbols_data.
                        get_exchange_symbol_data(self.position.symbol).prices_manager.mark_price
                )

            self.has_terminated = True

    async def terminate(self):
        self.sync_terminate()

    def is_pending(self) -> bool:
        return self.state is enums.States.OPENING



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_trading.enums import TraderOrderType
from tests.personal_data import DEFAULT_SYMBOL_QUANTITY, DEFAULT_ORDER_SYMBOL
from tests.test_utils.random_numbers import decimal_random_price, random_price, decimal_random_quantity, \
    random_recent_trade
import octobot_trading.personal_data.orders.groups as order_groups

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import stop_loss_sell_order, sell_limit_order

pytestmark = pytest.mark.asyncio


async def test_stop_loss_and_limit(stop_loss_sell_order, sell_limit_order):
    # fill both orders: stop loss first
    limit_order_price = decimal_random_price()
    quantity = decimal_random_quantity(max_value=DEFAULT_SYMBOL_QUANTITY)
    sell_limit_order.update(
        price=limit_order_price,
        quantity=quantity,
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.SELL_LIMIT,
    )
    stop_order_price = decimal_random_price(max_value=limit_order_price-1)
    stop_loss_sell_order.update(
        price=stop_order_price,
        quantity=quantity,
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.STOP_LOSS
    )
    oco_group = stop_loss_sell_order.exchange_manager.exchange_personal_data.orders_manager\
        .create_group(order_groups.OneCancelsTheOtherOrderGroup)
    stop_loss_sell_order.add_to_order_group(oco_group)
    sell_limit_order.add_to_order_group(oco_group)
    stop_loss_sell_order.exchange_manager.is_backtesting = True  # force update_order_status
    # initialize stop loss first
    await stop_loss_sell_order.initialize()
    await sell_limit_order.initialize()
    price_events_manager = stop_loss_sell_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    # stop loss sell order triggers when price is bellow or equal to its trigger price
    # sell limit order triggers when price is above or equal to its trigger price
    # here trigger both: stop order is triggered first (initialized first): sell limit order should be
    # cancelled and not filled even though its price has been hit
    price_events_manager.handle_recent_trades(
         [
            random_recent_trade(price=random_price(max_value=float(stop_order_price - 1)),
                                timestamp=stop_loss_sell_order.timestamp),
            random_recent_trade(price=random_price(min_value=float(limit_order_price + 1)),
                                timestamp=sell_limit_order.timestamp)
         ]
    )
    await wait_asyncio_next_cycle()
    assert stop_loss_sell_order.is_filled()
    assert sell_limit_order.is_cancelled()


async def test_limit_and_stop_loss(stop_loss_sell_order, sell_limit_order):
    # fill both orders: limit first
    limit_order_price = decimal_random_price()
    quantity = decimal_random_quantity(max_value=DEFAULT_SYMBOL_QUANTITY)
    sell_limit_order.update(
        price=limit_order_price,
        quantity=quantity,
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.SELL_LIMIT,
    )
    stop_order_price = decimal_random_price(max_value=limit_order_price-1)
    stop_loss_sell_order.update(
        price=stop_order_price,
        quantity=quantity,
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=TraderOrderType.STOP_LOSS
    )
    oco_group = stop_loss_sell_order.exchange_manager.exchange_personal_data.orders_manager\
        .create_group(order_groups.OneCancelsTheOtherOrderGroup)
    stop_loss_sell_order.add_to_order_group(oco_group)
    sell_limit_order.add_to_order_group(oco_group)
    stop_loss_sell_order.exchange_manager.is_backtesting = True  # force update_order_status
    # initialize limit order first
    await sell_limit_order.initialize()
    await stop_loss_sell_order.initialize()
    price_events_manager = stop_loss_sell_order.exchange_manager.exchange_symbols_data.get_exchange_symbol_data(
        DEFAULT_ORDER_SYMBOL).price_events_manager
    # stop loss sell order triggers when price is bellow or equal to its trigger price
    # sell limit order triggers when price is above or equal to its trigger price
    # here trigger both: limit is triggered first (initialized first): sell stop loss order should be
    # cancelled and not filled even though its price has been hit
    price_events_manager.handle_recent_trades(
         [
            random_recent_trade(price=random_price(max_value=float(stop_order_price - 1)),
                                timestamp=sell_limit_order.timestamp),
            random_recent_trade(price=random_price(min_value=float(limit_order_price + 1)),
                                timestamp=stop_loss_sell_order.timestamp)
         ]
    )
    await wait_asyncio_next_cycle()
    assert stop_loss_sell_order.is_cancelled()
    assert sell_limit_order.is_filled()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os

import mock
import pytest
from mock import patch, Mock, AsyncMock
import octobot_commons.constants as commons_constants
import octobot_trading.personal_data as personal_data
from octobot_trading.personal_data.orders import BuyMarketOrder, BuyLimitOrder

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
from tests.personal_data import DEFAULT_MARKET_QUANTITY
from tests.test_utils.random_numbers import decimal_random_price, decimal_random_quantity

pytestmark = pytest.mark.asyncio


async def test_handle_balance_update(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    if os.getenv('CYTHON_IGNORE'):
        return

    with patch.object(portfolio_manager.portfolio, 'update_portfolio_from_balance',
                      new=Mock()) as update_portfolio_from_balance_mock:
        update_portfolio_from_balance_mock.assert_not_called()

        portfolio_manager.handle_balance_update(None)
        update_portfolio_from_balance_mock.assert_not_called()

        trader.is_enabled = False
        portfolio_manager.handle_balance_update({})
        update_portfolio_from_balance_mock.assert_not_called()

        trader.is_enabled = True
        portfolio_manager.handle_balance_update({})
        update_portfolio_from_balance_mock.assert_called_once()


async def test_handle_balance_update_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    trader.simulate = False
    order = BuyMarketOrder(trader)

    if os.getenv('CYTHON_IGNORE'):
        return

    with patch.object(portfolio_manager, '_refresh_real_trader_portfolio',
                      new=AsyncMock()) as _refresh_real_trader_portfolio_mock, \
        patch.object(portfolio_manager, '_refresh_simulated_trader_portfolio_from_order',
                     new=Mock()) as _refresh_simulated_trader_portfolio_from_order_mock:
        _refresh_real_trader_portfolio_mock.assert_not_called()
        await portfolio_manager.handle_balance_update_from_order(order, True)
        _refresh_real_trader_portfolio_mock.assert_called_once()
        _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        _refresh_real_trader_portfolio_mock.reset_mock()
        await portfolio_manager.handle_balance_update_from_order(order, False)
        _refresh_real_trader_portfolio_mock.assert_not_called()
        _refresh_simulated_trader_portfolio_from_order_mock.assert_called_once()

    trader.simulate = True
    with patch.object(portfolio_manager, '_refresh_simulated_trader_portfolio_from_order',
                      new=Mock()) as _refresh_simulated_trader_portfolio_from_order_mock:
        _refresh_simulated_trader_portfolio_from_order_mock.assert_not_called()
        await portfolio_manager.handle_balance_update_from_order(order, True)
        _refresh_simulated_trader_portfolio_from_order_mock.assert_called_once()
        _refresh_simulated_trader_portfolio_from_order_mock.reset_mock()
        # ensure no side effect with require_exchange_update param
        await portfolio_manager.handle_balance_update_from_order(order, False)
        _refresh_simulated_trader_portfolio_from_order_mock.assert_called_once()

    trader.is_enabled = False
    trader.simulate = False
    assert not await portfolio_manager.handle_balance_update_from_order(order, True)
    assert not await portfolio_manager.handle_balance_update_from_order(order, False)


async def test_refresh_simulated_trader_portfolio_from_order(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    if os.getenv('CYTHON_IGNORE'):
        return
    order = BuyLimitOrder(trader)
    await order.initialize()
    with patch.object(portfolio_manager.portfolio, 'update_portfolio_available',
                      new=Mock()) as update_portfolio_available_mock:
        update_portfolio_available_mock.assert_not_called()
        portfolio_manager._refresh_simulated_trader_portfolio_from_order(order)
        update_portfolio_available_mock.assert_called_once()

    price = decimal_random_price()
    order.update(
        price=decimal_random_price(),
        quantity=decimal_random_quantity(max_value=DEFAULT_MARKET_QUANTITY / price),
        symbol="BTC/USDT"
    )
    await order.on_fill(force_fill=True)
    assert order.is_filled()

    with patch.object(portfolio_manager.portfolio, 'update_portfolio_from_filled_order',
                      new=Mock()) as update_portfolio_from_filled_order_mock:
        update_portfolio_from_filled_order_mock.assert_not_called()
        portfolio_manager._refresh_simulated_trader_portfolio_from_order(order)
        update_portfolio_from_filled_order_mock.assert_called_once()


async def test_load_simulated_portfolio_from_history(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager

    portfolio_manager.historical_portfolio_value_manager = mock.Mock(
        historical_ending_portfolio={
            "BTC": {
                commons_constants.PORTFOLIO_AVAILABLE: 1,
                commons_constants.PORTFOLIO_TOTAL: 10.11,
            },
            "ETH": {
                commons_constants.PORTFOLIO_AVAILABLE: -1,
                commons_constants.PORTFOLIO_TOTAL: 10,
            },
            "USDT": {
                commons_constants.PORTFOLIO_AVAILABLE: 34,
                commons_constants.PORTFOLIO_TOTAL: 34,
            }
        },
        stop=mock.AsyncMock()
    )
    portfolio_manager._load_simulated_portfolio_from_history()
    # ensure only the total value is loaded in simulated portfolio
    assert portfolio_manager.portfolio.portfolio == {
        "BTC": personal_data.SpotAsset("BTC", decimal.Decimal("10.11"), decimal.Decimal("10.11")),
        "ETH": personal_data.SpotAsset("ETH", decimal.Decimal("10"), decimal.Decimal("10")),
        "USDT": personal_data.SpotAsset("USDT", decimal.Decimal("34"), decimal.Decimal("34"))
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import contextlib
import os
import mock
import pytest
import pytest_asyncio

import octobot_commons.constants as commons_constants
from octobot_backtesting.backtesting import Backtesting
from octobot_backtesting.constants import CONFIG_BACKTESTING
import octobot_backtesting.time as backtesting_time
from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_commons.enums import TimeFrames

from octobot_commons.tests.test_config import load_test_config
from octobot_trading.api.exchange import create_exchange_builder, cancel_ccxt_throttle_task
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.exchanges.implementations.default_rest_exchange import DefaultRestExchange
from octobot_trading.exchanges.connectors.ccxt.ccxt_connector import CCXTConnector
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
import octobot_trading.personal_data as personal_data
from octobot_trading.enums import FeePropertyColumns, ExchangeConstantsMarketPropertyColumns, \
    ExchangeConstantsMarketPropertyColumns

import tests.exchanges.connectors.ccxt.mock_exchanges_data as mock_exchanges_data

pytestmark = pytest.mark.asyncio

TESTS_FOLDER = "tests"
TESTS_STATIC_FOLDER = os.path.join(TESTS_FOLDER, "static")
DEFAULT_EXCHANGE_NAME = "binanceus"
DEFAULT_FUTURE_EXCHANGE_NAME = "bybit"


class MockedCCXTConnector(CCXTConnector):
    @classmethod
    def get_name(cls):
        return DEFAULT_EXCHANGE_NAME

    async def load_symbol_markets(self, reload=False, forced_markets=None):
        if forced_markets is None:
            forced_markets = mock_exchanges_data.MOCKED_EXCHANGE_INFO.get(self.exchange_manager.exchange_name, None)
        await super().load_symbol_markets(
            reload=reload,
            forced_markets=forced_markets
        )

    def _should_authenticate(self):
        return False


class MockedRestExchange(DefaultRestExchange):
    DEFAULT_CONNECTOR_CLASS = MockedCCXTConnector

    @classmethod
    def get_exchange_connector_class(cls, exchange_manager):
        return cls.DEFAULT_CONNECTOR_CLASS

    @classmethod
    def is_default_exchange(cls) -> bool:
        return False


class MockedAutoFillRestExchange(MockedRestExchange):
    HAS_FETCHED_DETAILS = True
    _SUPPORTED_EXCHANGE_MOCK = []
    _EXCHANGE_DETAILS_MOCK = {}

    def _fetch_details(self, config, exchange_manager):
        pass

    @staticmethod
    def supported_autofill_exchanges(tentacle_config):
        return MockedAutoFillRestExchange._SUPPORTED_EXCHANGE_MOCK

    @classmethod
    async def get_autofilled_exchange_details(cls, aiohttp_session, tentacle_config, exchange_name):
        return MockedAutoFillRestExchange._EXCHANGE_DETAILS_MOCK[exchange_name]

    @classmethod
    @contextlib.contextmanager
    def patched_supported_exchanges(cls, exchange_details_by_exchange_name):
        previous_exchanges = cls._SUPPORTED_EXCHANGE_MOCK
        previous_details = cls._EXCHANGE_DETAILS_MOCK
        try:
            cls._SUPPORTED_EXCHANGE_MOCK = list(exchange_details_by_exchange_name)
            cls._EXCHANGE_DETAILS_MOCK = exchange_details_by_exchange_name
            yield
        finally:
            cls._SUPPORTED_EXCHANGE_MOCK = previous_exchanges
            cls._EXCHANGE_DETAILS_MOCK = previous_details


@pytest_asyncio.fixture
async def exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = True
    exchange_manager_instance.is_simulated = False
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        cancel_ccxt_throttle_task()
        await exchange_manager_instance.stop()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()


@pytest_asyncio.fixture
async def simulated_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = True
    exchange_manager_instance.is_simulated = True
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        cancel_ccxt_throttle_task()
        await exchange_manager_instance.stop()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()


@pytest_asyncio.fixture
async def margin_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_margin = True
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        cancel_ccxt_throttle_task()
        await exchange_manager_instance.stop()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()


@pytest_asyncio.fixture
async def margin_simulated_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_simulated = True
    exchange_manager_instance.is_margin = True
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        cancel_ccxt_throttle_task()
        await exchange_manager_instance.stop()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()


@pytest_asyncio.fixture
async def future_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_FUTURE_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_future = True
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        cancel_ccxt_throttle_task()
        await exchange_manager_instance.stop()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()


@pytest_asyncio.fixture
async def future_simulated_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_FUTURE_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_simulated = True
    exchange_manager_instance.is_future = True
    await exchange_manager_instance.initialize()
    try:
        yield exchange_manager_instance
    finally:
        cancel_ccxt_throttle_task()
        await exchange_manager_instance.stop()
        # let updaters gracefully shutdown
        await wait_asyncio_next_cycle()


@pytest_asyncio.fixture
async def exchange_builder(request):
    config = None
    exchange_name = DEFAULT_EXCHANGE_NAME
    if hasattr(request, "param"):
        config, exchange_name = request.param

    exchange_builder_instance = create_exchange_builder(config if config is not None else load_test_config(),
                                                        exchange_name).is_simulated().is_rest_only()
    try:
        yield exchange_builder_instance
    finally:
        await exchange_builder_instance.exchange_manager.stop()


# SIMULATED / BACKTESTING
DEFAULT_BACKTESTING_SYMBOL = "BTC/USDT"
DEFAULT_BACKTESTING_SPLIT_SYMBOL = "BTC", "USDT"
DEFAULT_BACKTESTING_CURRENCY = "BTC"
DEFAULT_BACKTESTING_MARKET = "USDT"
DEFAULT_BACKTESTING_TF = TimeFrames.ONE_HOUR


@pytest_asyncio.fixture
async def backtesting_config(request):
    config = load_test_config()
    config[CONFIG_BACKTESTING] = {}
    config[CONFIG_BACKTESTING][commons_constants.CONFIG_ENABLED_OPTION] = True
    if hasattr(request, "param"):
        ref_market = request.param
        config[commons_constants.CONFIG_TRADING][commons_constants.CONFIG_TRADER_REFERENCE_MARKET] = ref_market
    return config


@pytest_asyncio.fixture
async def fake_backtesting(backtesting_config):
    return Backtesting(config=backtesting_config,
                       exchange_ids=[],
                       matrix_id="",
                       backtesting_files=[])


@pytest_asyncio.fixture
async def backtesting_exchange_manager(request, backtesting_config, fake_backtesting):
    config = None
    exchange_name = DEFAULT_EXCHANGE_NAME
    is_spot = True
    is_margin = False
    is_future = False
    if hasattr(request, "param"):
        config, exchange_name, is_spot, is_margin, is_future = request.param

    if config is None:
        config = backtesting_config
    exchange_manager_instance = ExchangeManager(config, exchange_name)
    exchange_manager_instance.is_backtesting = True
    exchange_manager_instance.is_spot_only = is_spot
    exchange_manager_instance.is_margin = is_margin
    exchange_manager_instance.is_future = is_future
    exchange_manager_instance.backtesting = fake_backtesting
    exchange_manager_instance.backtesting.time_manager = backtesting_time.TimeManager(config)
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    await exchange_manager_instance.stop()


@pytest_asyncio.fixture
async def simulated_trader(simulated_exchange_manager):
    config = load_test_config()
    trader_instance = TraderSimulator(config, simulated_exchange_manager)
    await trader_instance.initialize()
    return config, simulated_exchange_manager, trader_instance


@pytest_asyncio.fixture
async def backtesting_trader(backtesting_config, backtesting_exchange_manager):
    trader_instance = TraderSimulator(backtesting_config, backtesting_exchange_manager)
    await trader_instance.initialize()
    return backtesting_config, backtesting_exchange_manager, trader_instance


def storage_mock():
    return mock.Mock(
        portfolio_storage=mock.Mock(
            get_db=mock.Mock(
                return_value=mock.Mock(
                    all=mock.AsyncMock(return_value=[])
                )
            ),
            save_historical_portfolio_value=mock.AsyncMock(),
        ),
        stop=mock.AsyncMock(),
    )


def get_fees_mock(symbol, rate=0.1, cost=0.1):  # huge fees for tests
    return {
        FeePropertyColumns.RATE.value: rate,
        FeePropertyColumns.COST.value: cost,
        FeePropertyColumns.CURRENCY.value: symbol
    }


def set_future_exchange_fees(exchange, symbol, taker=0.0004, maker=0.0004):
    exchange.client.markets[symbol]['taker'] = taker
    exchange.client.markets[symbol]['maker'] = maker


@pytest_asyncio.fixture
async def backtesting_trader_with_historical_pf_value_manager(backtesting_trader):
    backtesting_config, backtesting_exchange_manager, trader_instance = backtesting_trader
    backtesting_exchange_manager.storage_manager = storage_mock()
    historical_portfolio_value_manager_inst = personal_data.HistoricalPortfolioValueManager(
        backtesting_exchange_manager.exchange_personal_data.portfolio_manager
    )
    backtesting_exchange_manager.exchange_personal_data.portfolio_manager.historical_portfolio_value_manager = \
        historical_portfolio_value_manager_inst
    await historical_portfolio_value_manager_inst.initialize()
    backtesting_exchange_manager.exchange_personal_data.portfolio_manager.handle_profitability_recalculation(True)
    return backtesting_config, backtesting_exchange_manager, trader_instance


@pytest_asyncio.fixture
async def margin_backtesting_trader(backtesting_config, backtesting_exchange_manager):
    trader_instance = TraderSimulator(backtesting_config, backtesting_exchange_manager)
    await trader_instance.initialize()
    return backtesting_config, backtesting_exchange_manager, trader_instance


@pytest_asyncio.fixture
async def future_backtesting_trader(backtesting_config, backtesting_exchange_manager):
    trader_instance = TraderSimulator(backtesting_config, backtesting_exchange_manager)
    await trader_instance.initialize()
    return backtesting_config, backtesting_exchange_manager, trader_instance



#  Drakkar-Software OctoBot-Private-Tentacles
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.enums as common_enums
import octobot_commons.constants as commons_constants
import octobot_trading.enums as enums
import octobot_trading.api as trading_api
import octobot_trading.exchanges as exchanges
import octobot_tentacles_manager.api as tentacles_manager_api

import octobot_trading.personal_data as personal_data

import octobot_trading.util.test_tools.exchange_data as exchange_data_import


async def create_test_exchange_manager(
        config: dict,
        exchange_name: str,
        is_spot_only: bool = True,
        is_margin: bool = False,
        is_future: bool = False,
        rest_only: bool = False,
        is_real: bool = True,
        is_sandboxed: bool = False,
        ignore_exchange_config: bool = True) -> exchanges.ExchangeManager:
    if ignore_exchange_config:
        # enable exchange name in config
        config[commons_constants.CONFIG_EXCHANGES][exchange_name] = {commons_constants.CONFIG_ENABLED_OPTION: True}

    builder = exchanges.create_exchange_builder_instance(config, exchange_name)
    builder.disable_trading_mode()
    builder.use_tentacles_setup_config(tentacles_manager_api.get_tentacles_setup_config())
    if is_spot_only:
        builder.is_spot_only()
    if is_margin:
        builder.is_margin()
    if is_future:
        builder.is_future()
    if rest_only:
        builder.is_rest_only()
    if is_sandboxed:
        builder.is_sandboxed(is_sandboxed)
    if is_real:
        builder.is_real()
    else:
        builder.is_simulated()
    await builder.build()
    return builder.exchange_manager


async def stop_test_exchange_manager(exchange_manager_instance: exchanges.ExchangeManager):
    trading_api.cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await asyncio_tools.wait_asyncio_next_cycle()


async def add_symbols_details(
    exchange_manager, symbols: list, time_frame: str, exchange_data: exchange_data_import.ExchangeData, history_size=1
) -> exchange_data_import.ExchangeData:
    parsed_tf = common_enums.TimeFrames(time_frame)

    async def _update_ohlcv(symbol):
        market = exchange_manager.exchange.connector.client.markets[symbol]
        ohlcvs = await exchange_manager.exchange.get_symbol_prices(symbol, parsed_tf, limit=history_size)
        details = exchange_data_import.MarketDetails(
            id=market[enums.ExchangeConstantsMarketStatusColumns.ID.value],
            symbol=market[enums.ExchangeConstantsMarketStatusColumns.SYMBOL.value],
            info=market[enums.ExchangeConstantsMarketStatusColumns.INFO.value],
            time_frame=time_frame,
            close=[ohlcv[common_enums.PriceIndexes.IND_PRICE_CLOSE.value] for ohlcv in ohlcvs],
            open=[ohlcv[common_enums.PriceIndexes.IND_PRICE_OPEN.value] for ohlcv in ohlcvs],
            high=[ohlcv[common_enums.PriceIndexes.IND_PRICE_HIGH.value] for ohlcv in ohlcvs],
            low=[ohlcv[common_enums.PriceIndexes.IND_PRICE_LOW.value] for ohlcv in ohlcvs],
            volume=[ohlcv[common_enums.PriceIndexes.IND_PRICE_VOL.value] for ohlcv in ohlcvs],
            time=[ohlcv[common_enums.PriceIndexes.IND_PRICE_TIME.value] for ohlcv in ohlcvs],
        )
        exchange_data.markets.append(details)

    await exchange_manager.exchange.connector.load_symbol_markets()
    await asyncio.gather(*(_update_ohlcv(symbol) for symbol in symbols))
    return exchange_data


async def get_portfolio(exchange_manager, as_float=False) -> dict:
    balance = await exchange_manager.exchange.get_balance()
    # filter out 0 values
    return {
        asset: {key: float(val) if as_float else val for key, val in values.items()}  # use float for values
        for asset, values in balance.items()
        if any(value for value in values.values())
    }


async def get_open_orders(exchange_manager, exchange_data: exchange_data_import.ExchangeData) -> list:
    open_orders = []

    async def _get_orders(symbol):
        orders = await exchange_manager.exchange.get_open_orders(symbol=symbol)
        open_orders.extend(
            personal_data.create_order_instance_from_raw(
                exchange_manager.trader, order, force_open_or_pending_creation=True
            ).to_dict()
            for order in orders
        )

    await asyncio.gather(*(_get_orders(market.symbol) for market in exchange_data.markets))
    return open_orders


async def get_trades(exchange_manager, exchange_data: exchange_data_import.ExchangeData) -> list:
    trades = []

    async def _get_trades(symbol):
        row_trades = await exchange_manager.exchange.get_my_recent_trades(symbol=symbol)
        trades.extend(
            personal_data.create_trade_instance_from_raw(exchange_manager.trader, raw_trade).to_dict()
            for raw_trade in row_trades
        )

    await asyncio.gather(*(_get_trades(market.symbol) for market in exchange_data.markets))
    return trades


async def create_orders(exchange_manager, exchange_data: exchange_data_import.ExchangeData, orders: list) -> list:
    async def _create_order(order_dict) -> personal_data.Order:
        symbol = order_dict[enums.ExchangeConstantsOrderColumns.SYMBOL.value]
        side, order_type = personal_data.parse_order_type(order_dict)
        created_order = await exchange_manager.exchange.create_order(
            order_type,
            symbol,
            decimal.Decimal(str(order_dict[enums.ExchangeConstantsOrderColumns.AMOUNT.value])),
            price=decimal.Decimal(str(order_dict[enums.ExchangeConstantsOrderColumns.PRICE.value])),
            side=side,
            current_price=decimal.Decimal(str(exchange_data.get_price(symbol))),
            reduce_only=order_dict[enums.ExchangeConstantsOrderColumns.REDUCE_ONLY.value],
        )
        # created_order = order_dict
        # is private, to use in tests context only
        return personal_data.create_order_instance_from_raw(
            exchange_manager.trader, created_order, force_open_or_pending_creation=True
        )

    return await asyncio.gather(*(_create_order(order_dict) for order_dict in orders))
